import enum, os, re, struct, sys
from typing import Tuple

from .log import log


PORTS = range(4)


def _indent(s):
    return re.sub(r'^', '    ', s, flags=re.MULTILINE)


def _format_collection(coll, delim_open, delim_close):
    elements = [_format(x) for x in coll]
    if elements and '\n' in elements[0]:
        return delim_open + '\n' + ',\n'.join(_indent(e) for e in elements) + delim_close
    else:
        return delim_open + ', '.join(elements) + delim_close


def _format(obj):
    if isinstance(obj, float):
        return '%.02f' % obj
    elif isinstance(obj, tuple):
        return _format_collection(obj, '(', ')')
    elif isinstance(obj, list):
        return _format_collection(obj, '[', ']')
    elif isinstance(obj, enum.Enum):
        return repr(obj)
    else:
        return str(obj)


def try_enum(enum, val):
    try:
        return enum(val)
    except ValueError:
        log.info('unknown %s: %s' % (enum.__name__, val))
        return val


def unpack(fmt, stream):
    fmt = '>' + fmt
    size = struct.calcsize(fmt)
    bytes = stream.read(size)
    if not bytes:
        raise EOFError()
    return struct.unpack(fmt, bytes)


def expect_bytes(expected_bytes, stream):
    read_bytes = stream.read(len(expected_bytes))
    if read_bytes != expected_bytes:
        raise Exception(f'expected {expected_bytes}, but got: {read_bytes}')


class Base:
    __slots__: Tuple = ()

    def _attr_repr(self, attr):
        return attr + '=' + _format(getattr(self, attr))

    def __repr__(self):
        attrs = []
        for attr in dir(self):
            # uppercase names are nested classes
            if not (attr.startswith('_') or attr[0].isupper()):
                s = self._attr_repr(attr)
                if s:
                    attrs.append(_indent(s))

        return '%s(\n%s)' % (self.__class__.__name__, ',\n'.join(attrs))


class Enum(enum.Enum):
    def __repr__(self):
        return '%r:%s' % (self._value_, self._name_)


class IntEnum(enum.IntEnum):
    def __repr__(self):
        return '%d:%s' % (self._value_, self._name_)

    @classmethod
    def _missing_(cls, value):
        val_desc = f'0x{value:x}' if isinstance(value, int) else f'{value}'
        raise ValueError(f'{val_desc} is not a valid {cls.__name__}') from None


class IntFlag(enum.IntFlag):
    def __repr__(self):
        members, _ = enum._decompose(self.__class__, self._value_)
        return '%s:%s' % (bin(self._value_), '|'.join([str(m._name_ or m._value_) for m in members]))


class EOFError(IOError):
    def __init__(self):
        super().__init__('unexpected end of file')



from __future__ import annotations

from typing import Optional, Sequence, Tuple, Union

from . import id as sid
from .util import *

# The first frame of the game is indexed -123, counting up to zero (which is when the word "GO" appears). But since players actually get control before frame zero (!!!), we need to record these frames.
FIRST_FRAME_INDEX = -123


class EventType(IntEnum):
    """Slippi events that can appear in a game's `raw` data."""

    EVENT_PAYLOADS = 0x35
    GAME_START = 0x36
    FRAME_PRE = 0x37
    FRAME_POST = 0x38
    GAME_END = 0x39
    FRAME_START = 0x3A
    ITEM = 0x3B
    FRAME_END = 0x3C


class Start(Base):
    """Information used to initialize the game such as the game mode, settings, characters & stage."""

    is_teams: bool #: True if this was a teams game
    players: Tuple[Optional[Start.Player]] #: Players in this game by port (port 1 is at index 0; empty ports will contain None)
    random_seed: int #: Random seed before the game start
    slippi: Start.Slippi #: Information about the Slippi recorder that generated this replay
    stage: sid.Stage #: Stage on which this game was played
    is_pal: Optional[bool] #: `added(1.5.0)` True if this was a PAL version of Melee
    is_frozen_ps: Optional[bool] #: `added(2.0.0)` True if frozen Pokemon Stadium was enabled

    def __init__(self, is_teams: bool, players: Tuple[Optional[Start.Player]], random_seed: int, slippi: Start.Slippi, stage: sid.Stage, is_pal: Optional[bool] = None, is_frozen_ps: Optional[bool] = None):
        self.is_teams = is_teams
        self.players = players
        self.random_seed = random_seed
        self.slippi = slippi
        self.stage = stage
        self.is_pal = is_pal
        self.is_frozen_ps = is_frozen_ps

    @classmethod
    def _parse(cls, stream):
        slippi_ = cls.Slippi._parse(stream)

        stream.read(8)
        (is_teams,) = unpack('?', stream)

        stream.read(5)
        (stage,) = unpack('H', stream)
        stage = sid.Stage(stage)

        stream.read(80)
        players = []
        for i in PORTS:
            (character, type, stocks, costume) = unpack('BBBB', stream)

            stream.read(5)
            (team,) = unpack('B', stream)
            stream.read(26)

            try: type = cls.Player.Type(type)
            except ValueError: type = None

            if type is not None:
                character = sid.CSSCharacter(character)
                team = cls.Player.Team(team) if is_teams else None
                player = cls.Player(character=character, type=type, stocks=stocks, costume=costume, team=team)
            else:
                player = None

            players.append(player)

        stream.read(72)
        (random_seed,) = unpack('L', stream)

        try: # v1.0.0
            for i in PORTS:
                (dash_back, shield_drop) = unpack('LL', stream)
                dash_back = cls.Player.UCF.DashBack(dash_back)
                shield_drop = cls.Player.UCF.ShieldDrop(shield_drop)
                if players[i]:
                    players[i].ucf = cls.Player.UCF(dash_back, shield_drop)
        except EOFError: pass

        try: # v1.3.0
            for i in PORTS:
                tag_bytes = stream.read(16)
                if players[i]:
                    try:
                        null_pos = tag_bytes.index(0)
                        tag_bytes = tag_bytes[:null_pos]
                    except ValueError: pass
                    players[i].tag = tag_bytes.decode('shift-jis').rstrip()
        except EOFError: pass

        # v1.5.0
        try: (is_pal,) = unpack('?', stream)
        except EOFError: is_pal = None

        # v2.0.0
        try: (is_frozen_ps,) = unpack('?', stream)
        except EOFError: is_frozen_ps = None

        return cls(
            is_teams=is_teams,
            players=tuple(players),
            random_seed=random_seed,
            slippi=slippi_,
            stage=stage,
            is_pal=is_pal,
            is_frozen_ps=is_frozen_ps)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.is_teams == other.is_teams and self.players == other.players and self.random_seed == other.random_seed and self.slippi == other.slippi and self.stage is other.stage


    class Slippi(Base):
        """Information about the Slippi recorder that generated this replay."""

        version: Start.Slippi.Version #: Slippi version number

        def __init__(self, version: Start.Slippi.Version):
            self.version = version

        @classmethod
        def _parse(cls, stream):
            return cls(cls.Version(*unpack('BBBB', stream)))

        def __eq__(self, other):
            if not isinstance(other, self.__class__):
                return NotImplemented
            return self.version == other.version


        class Version(Base):

            major: int
            minor: int
            revision: int

            def __init__(self, major: int, minor: int, revision: int, build = None):
                self.major = major
                self.minor = minor
                self.revision = revision
                # build was obsoleted in 2.0.0, and never held a nonzero value

            def __repr__(self):
                return '%d.%d.%d' % (self.major, self.minor, self.revision)

            def __eq__(self, other):
                if not isinstance(other, self.__class__):
                    return NotImplemented
                return self.major == other.major and self.minor == other.minor and self.revision == other.revision


    class Player(Base):
        character: sid.CSSCharacter #: Character selected
        type: Start.Player.Type #: Player type (human/cpu)
        stocks: int #: Starting stock count
        costume: int #: Costume ID
        team: Optional[Start.Player.Team] #: Team, if this was a teams game
        ucf: Start.Player.UCF #: UCF feature toggles
        tag: Optional[str] #: Name tag

        def __init__(self, character: sid.CSSCharacter, type: Start.Player.Type, stocks: int, costume: int, team: Optional[Start.Player.Team], ucf: Start.Player.UCF = None, tag: Optional[str] = None):
            self.character = character
            self.type = type
            self.stocks = stocks
            self.costume = costume
            self.team = team
            self.ucf = ucf or self.UCF()
            self.tag = tag

        def __eq__(self, other):
            if not isinstance(other, self.__class__):
                return NotImplemented
            return self.character is other.character and self.type is other.type and self.stocks == other.stocks and self.costume == other.costume and self.team is other.team and self.ucf == other.ucf


        class Type(IntEnum):
            HUMAN = 0
            CPU = 1


        class Team(IntEnum):
            RED = 0
            BLUE = 1
            GREEN = 2


        class UCF(Base):
            dash_back: Start.Player.UCF.DashBack #: UCF dashback status
            shield_drop: Start.Player.UCF.ShieldDrop #: UCF shield drop status

            def __init__(self, dash_back: Start.Player.UCF.DashBack = None, shield_drop: Start.Player.UCF.ShieldDrop = None):
                self.dash_back = dash_back or self.DashBack.OFF
                self.shield_drop = shield_drop or self.ShieldDrop.OFF

            def __eq__(self, other):
                if not isinstance(other, self.__class__):
                    return NotImplemented
                return self.dash_back == other.dash_back and self.shield_drop == other.shield_drop


            class DashBack(IntEnum):
                OFF = 0
                UCF = 1
                ARDUINO = 2


            class ShieldDrop(IntEnum):
                OFF = 0
                UCF = 1
                ARDUINO = 2


class End(Base):
    """Information about the end of the game."""

    method: End.Method #: `changed(2.0.0)` How the game ended
    lras_initiator: Optional[int] #: `added(2.0.0)` Index of player that LRAS'd, if any

    def __init__(self, method: End.Method, lras_initiator: Optional[int] = None):
        self.method = method
        self.lras_initiator = lras_initiator

    @classmethod
    def _parse(cls, stream):
        (method,) = unpack('B', stream)
        try: # v2.0.0
            (lras,) = unpack('B', stream)
            lras_initiator = lras if lras < len(PORTS) else None
        except EOFError:
            lras_initiator = None
        return cls(cls.Method(method), lras_initiator)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.method is other.method


    class Method(IntEnum):
        INCONCLUSIVE = 0 # `obsoleted(2.0.0)`
        TIME = 1 # `added(2.0.0)`
        GAME = 2 # `added(2.0.0)`
        CONCLUSIVE = 3 # `obsoleted(2.0.0)`
        NO_CONTEST = 7 # `added(2.0.0)`


class Frame(Base):
    """A single frame of the game. Includes data for all characters."""

    __slots__ = 'index', 'ports', 'items', 'start', 'end'

    index: int
    ports: Sequence[Optional[Frame.Port]] #: Frame data for each port (port 1 is at index 0; empty ports will contain None)
    items: Sequence[Frame.Item] #: `added(3.0.0)` Active items (includes projectiles)
    start: Optional[Frame.Start] #: `added(2.2.0)` Start-of-frame data
    end: Optional[Frame.End] #: `added(2.2.0)` End-of-frame data

    def __init__(self, index: int):
        self.index = index
        self.ports = [None, None, None, None]
        self.items = []
        self.start = None
        self.end = None

    def _finalize(self):
        self.ports = tuple(self.ports)
        self.items = tuple(self.items)


    class Port(Base):
        """Frame data for a given port. Can include two characters' frame data (ICs)."""

        __slots__ = 'leader', 'follower'

        leader: Frame.Port.Data #: Frame data for the controlled character
        follower: Optional[Frame.Port.Data] #: Frame data for the follower (Nana), if any

        def __init__(self):
            self.leader = self.Data()
            self.follower = None


        class Data(Base):
            """Frame data for a given character. Includes both pre-frame and post-frame data."""

            __slots__ = '_pre', '_post'

            def __init__(self):
                self._pre = None
                self._post = None

            @property
            def pre(self) -> Optional[Frame.Port.Data.Pre]:
                """Pre-frame update data"""
                if self._pre and not isinstance(self._pre, self.Pre):
                    self._pre = self.Pre._parse(self._pre)
                return self._pre

            @property
            def post(self) -> Optional[Frame.Port.Data.Post]:
                """Post-frame update data"""
                if self._post and not isinstance(self._post, self.Post):
                    self._post = self.Post._parse(self._post)
                return self._post


            class Pre(Base):
                """Pre-frame update data, required to reconstruct a replay. Information is collected right before controller inputs are used to figure out the character's next action."""

                __slots__ = 'state', 'position', 'direction', 'joystick', 'cstick', 'triggers', 'buttons', 'random_seed', 'raw_analog_x', 'damage'

                state: Union[sid.ActionState, int]
                position: Position
                direction: Direction
                joystick: Position
                cstick: Position
                triggers: Triggers
                buttons: Buttons
                random_seed: int
                raw_analog_x: Optional[int]
                damage: Optional[float]

                def __init__(self, state: Union[sid.ActionState, int], position: Position, direction: Direction, joystick: Position, cstick: Position, triggers: Triggers, buttons: Buttons, random_seed: int, raw_analog_x: Optional[int] = None, damage: Optional[float] = None):
                    self.state = state #: :py:class:`slippi.id.ActionState` | int: Character's action state
                    self.position = position #: :py:class:`Position`: Character's position
                    self.direction = direction #: :py:class:`Direction`: Direction the character is facing
                    self.joystick = joystick #: :py:class:`Position`: Processed analog joystick position
                    self.cstick = cstick #: :py:class:`Position`: Processed analog c-stick position
                    self.triggers = triggers #: :py:class:`Triggers`: Trigger state
                    self.buttons = buttons #: :py:class:`Buttons`: Button state
                    self.random_seed = random_seed #: int: Random seed at this point
                    self.raw_analog_x = raw_analog_x #: int | None: `added(1.2.0)` Raw x analog controller input (for UCF)
                    self.damage = damage #: float | None: `added(1.4.0)` Current damage percent

                @classmethod
                def _parse(cls, stream):
                    (random_seed, state, position_x, position_y, direction, joystick_x, joystick_y, cstick_x, cstick_y, trigger_logical, buttons_logical, buttons_physical, trigger_physical_l, trigger_physical_r) = unpack('LHffffffffLHff', stream)

                    # v1.2.0
                    try: raw_analog_x = unpack('B', stream)
                    except EOFError: raw_analog_x = None

                    # v1.4.0
                    try: damage = unpack('f', stream)
                    except EOFError: damage = None

                    return cls(
                        state=try_enum(sid.ActionState, state),
                        position=Position(position_x, position_y),
                        direction=Direction(direction),
                        joystick=Position(joystick_x, joystick_y),
                        cstick=Position(cstick_x, cstick_y),
                        triggers=Triggers(trigger_logical, trigger_physical_l, trigger_physical_r),
                        buttons=Buttons(buttons_logical, buttons_physical),
                        random_seed=random_seed,
                        raw_analog_x=raw_analog_x,
                        damage=damage)


            class Post(Base):
                """Post-frame update data, for making decisions about game states (such as computing stats). Information is collected at the end of collision detection, which is the last consideration of the game engine."""

                __slots__ = 'character', 'state', 'position', 'direction', 'damage', 'shield', 'stocks', 'last_attack_landed', 'last_hit_by', 'combo_count', 'state_age', 'flags', 'hit_stun', 'airborne', 'ground', 'jumps', 'l_cancel'

                character: sid.InGameCharacter #: In-game character (can only change for Zelda/Sheik). Check on first frame to determine if Zelda started as Sheik
                state: Union[sid.ActionState, int] #: Character's action state
                position: Position #: Character's position
                direction: Direction #: Direction the character is facing
                damage: float #: Current damage percent
                shield: float #: Current size of shield
                stocks: int #: Number of stocks remaining
                last_attack_landed: Union[Attack, int] #: Last attack that this character landed
                last_hit_by: Optional[int] #: Port of character that last hit this character
                combo_count: int #: Combo count as defined by the game
                state_age: Optional[float] #: `added(0.2.0)` Number of frames action state has been active. Can have a fractional component for certain actions
                flags: Optional[StateFlags] #: `added(2.0.0)` State flags
                hit_stun: Optional[float] #: `added(2.0.0)` Number of hitstun frames remaining
                airborne: Optional[bool] #: `added(2.0.0)` True if character is airborne
                ground: Optional[int] #: `added(2.0.0)` ID of ground character is standing on, if any
                jumps: Optional[int] #: `added(2.0.0)` Jumps remaining
                l_cancel: Optional[LCancel] #: `added(2.0.0)` L-cancel status, if any

                def __init__(self, character: sid.InGameCharacter, state: Union[sid.ActionState, int], position: Position, direction: Direction, damage: float, shield: float, stocks: int, last_attack_landed: Union[Attack, int], last_hit_by: Optional[int], combo_count: int, state_age: Optional[float] = None, flags: Optional[StateFlags] = None, hit_stun: Optional[float] = None, airborne: Optional[bool] = None, ground: Optional[int] = None, jumps: Optional[int] = None, l_cancel: Optional[LCancel] = None):
                    self.character = character
                    self.state = state
                    self.position = position
                    self.direction = direction
                    self.damage = damage
                    self.shield = shield
                    self.stocks = stocks
                    self.last_attack_landed = last_attack_landed
                    self.last_hit_by = last_hit_by
                    self.combo_count = combo_count
                    self.state_age = state_age
                    self.flags = flags
                    self.hit_stun = hit_stun
                    self.airborne = airborne
                    self.ground = ground
                    self.jumps = jumps
                    self.l_cancel = l_cancel

                @classmethod
                def _parse(cls, stream):
                    (character, state, position_x, position_y, direction, damage, shield, last_attack_landed, combo_count, last_hit_by, stocks) = unpack('BHfffffBBBB', stream)

                    # v0.2.0
                    try: (state_age,) = unpack('f', stream)
                    except EOFError: state_age = None

                    try: # v2.0.0
                        flags = unpack('5B', stream)
                        (misc_as, airborne, maybe_ground, jumps, l_cancel) = unpack('f?HBB', stream)
                        flags = StateFlags(flags[0] +
                                           flags[1] * 2**8 +
                                           flags[2] * 2**16 +
                                           flags[3] * 2**24 +
                                           flags[4] * 2**32)
                        ground = maybe_ground if not airborne else None
                        hit_stun = misc_as if flags.HIT_STUN else None
                        l_cancel = LCancel(l_cancel) if l_cancel else None
                    except EOFError:
                        (flags, hit_stun, airborne, ground, jumps, l_cancel) = [None] * 6

                    return cls(
                        character=sid.InGameCharacter(character),
                        state=try_enum(sid.ActionState, state),
                        state_age=state_age,
                        position=Position(position_x, position_y),
                        direction=Direction(direction),
                        damage=damage,
                        shield=shield,
                        stocks=stocks,
                        last_attack_landed=try_enum(Attack, last_attack_landed) if last_attack_landed else None,
                        last_hit_by=last_hit_by if last_hit_by < 4 else None,
                        combo_count=combo_count,
                        flags=flags,
                        hit_stun=hit_stun,
                        airborne=airborne,
                        ground=ground,
                        jumps=jumps,
                        l_cancel=l_cancel)


    class Item(Base):
        """An active item (includes projectiles)."""

        __slots__ = 'type', 'state', 'direction', 'velocity', 'position', 'damage', 'timer', 'spawn_id'

        type: sid.Item #: Item type
        state: int #: Item's action state
        direction: Direction #: Direction item is facing
        velocity: Velocity #: Item's velocity
        position: Position #: Item's position
        damage: int #: Amount of damage item has taken
        timer: int #: Frames remaining until item expires
        spawn_id: int #: Unique ID per item spawned (0, 1, 2, ...)

        def __init__(self, type: sid.Item, state: int, direction: Direction, velocity: Velocity, position: Position, damage: int, timer: int, spawn_id: int):
            self.type = type
            self.state = state
            self.direction = direction
            self.velocity = velocity
            self.position = position
            self.damage = damage
            self.timer = timer
            self.spawn_id = spawn_id

        @classmethod
        def _parse(cls, stream):
            (type, state, direction, x_vel, y_vel, x_pos, y_pos, damage, timer, spawn_id) = unpack('HB5fHfI', stream)
            return cls(
                type=try_enum(sid.Item, type),
                state=state,
                direction=Direction(direction),
                velocity=Velocity(x_vel, y_vel),
                position=Position(x_pos, y_pos),
                damage=damage,
                timer=timer,
                spawn_id=spawn_id)

        def __eq__(self, other):
            if not isinstance(other, self.__class__):
                return NotImplemented
            return self.type == other.type and self.state == other.state and self.direction == other.direction and self.velocity == other.velocity and self.position == other.position and self.damage == other.damage and self.timer == other.timer and self.spawn_id == other.spawn_id


    class Start(Base):
        """Start-of-frame data."""

        __slots__ = 'random_seed'

        random_seed: int #: The random seed at the start of the frame

        def __init__(self, random_seed: int):
            self.random_seed = random_seed

        @classmethod
        def _parse(cls, stream):
            (random_seed,) = unpack('I', stream)
            random_seed = random_seed
            return cls(random_seed)

        def __eq__(self, other):
            if not isinstance(other, self.__class__):
                return NotImplemented
            return self.random_seed == other.random_seed


    class End(Base):
        """End-of-frame data."""

        def __init__(self):
            pass

        @classmethod
        def _parse(cls, stream):
            return cls()

        def __eq__(self, other):
            if not isinstance(other, self.__class__):
                return NotImplemented
            return True


    class Event(Base):
        """Temporary wrapper used while parsing frame data."""

        __slots__ = 'id', 'type', 'data'

        def __init__(self, id, type, data):
            self.id = id
            self.type = type
            self.data = data


        class Id(Base):
            __slots__ = 'frame'

            def __init__(self, stream):
                (self.frame,) = unpack('i', stream)


        class PortId(Id):
            __slots__ = 'port', 'is_follower'

            def __init__(self, stream):
                (self.frame, self.port, self.is_follower) = unpack('iB?', stream)


        class Type(Enum):
            START = 'start'
            END = 'end'
            PRE = 'pre'
            POST = 'post'
            ITEM = 'item'


class Position(Base):
    __slots__ = 'x', 'y'

    x: float
    y: float

    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.x == other.x and self.y == other.y

    def __repr__(self):
        return '(%.2f, %.2f)' % (self.x, self.y)


class Velocity(Base):
    __slots__ = 'x', 'y'

    x: float
    y: float

    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.x == other.x and self.y == other.y

    def __repr__(self):
        return '(%.2f, %.2f)' % (self.x, self.y)


class Direction(IntEnum):
    LEFT = -1
    RIGHT = 1


class Attack(IntEnum):
    NON_STALING = 1
    JAB_1 = 2
    JAB_2 = 3
    JAB_3 = 4
    RAPID_JABS = 5
    DASH_ATTACK = 6
    SIDE_TILT = 7
    UP_TILT = 8
    DOWN_TILT = 9
    SIDE_SMASH = 10
    UP_SMASH = 11
    DOWN_SMASH = 12
    NAIR = 13
    FAIR = 14
    BAIR = 15
    UAIR = 16
    DAIR = 17
    NEUTRAL_SPECIAL = 18
    SIDE_SPECIAL = 19
    UP_SPECIAL = 20
    DOWN_SPECIAL = 21
    KIRBY_HAT_MARIO_NEUTRAL_SPECIAL = 22
    KIRBY_HAT_FOX_NEUTRAL_SPECIAL = 23
    KIRBY_HAT_CFALCON_NEUTRAL_SPECIAL = 24
    KIRBY_HAT_DKNEUTRAL_SPECIAL = 25
    KIRBY_HAT_BOWSER_NEUTRAL_SPECIAL = 26
    KIRBY_HAT_LINK_NEUTRAL_SPECIAL = 27
    KIRBY_HAT_SHEIK_NEUTRAL_SPECIAL = 28
    KIRBY_HAT_NESS_NEUTRAL_SPECIAL = 29
    KIRBY_HAT_PEACH_NEUTRAL_SPECIAL = 30
    KIRBY_HAT_ICE_CLIMBER_NEUTRAL_SPECIAL = 31
    KIRBY_HAT_PIKACHU_NEUTRAL_SPECIAL = 32
    KIRBY_HAT_SAMUS_NEUTRAL_SPECIAL = 33
    KIRBY_HAT_YOSHI_NEUTRAL_SPECIAL = 34
    KIRBY_HAT_JIGGLYPUFF_NEUTRAL_SPECIAL = 35
    KIRBY_HAT_MEWTWO_NEUTRAL_SPECIAL = 36
    KIRBY_HAT_LUIGI_NEUTRAL_SPECIAL = 37
    KIRBY_HAT_MARTH_NEUTRAL_SPECIAL = 38
    KIRBY_HAT_ZELDA_NEUTRAL_SPECIAL = 39
    KIRBY_HAT_YOUNG_LINK_NEUTRAL_SPECIAL = 40
    KIRBY_HAT_DOC_NEUTRAL_SPECIAL = 41
    KIRBY_HAT_FALCO_NEUTRAL_SPECIAL = 42
    KIRBY_HAT_PICHU_NEUTRAL_SPECIAL = 43
    KIRBY_HAT_GAME_AND_WATCH_NEUTRAL_SPECIAL = 44
    KIRBY_HAT_GANON_NEUTRAL_SPECIAL = 45
    KIRBY_HAT_ROY_NEUTRAL_SPECIAL = 46
    GET_UP_ATTACK_FROM_BACK = 50
    GET_UP_ATTACK_FROM_FRONT = 51
    PUMMEL = 52
    FORWARD_THROW = 53
    BACK_THROW = 54
    UP_THROW = 55
    DOWN_THROW = 56
    CARGO_FORWARD_THROW = 57
    CARGO_BACK_THROW = 58
    CARGO_UP_THROW = 59
    CARGO_DOWN_THROW = 60
    LEDGE_GET_UP_ATTACK_100 = 61
    LEDGE_GET_UP_ATTACK = 62
    BEAM_SWORD_JAB = 63
    BEAM_SWORD_TILT_SWING = 64
    BEAM_SWORD_SMASH_SWING = 65
    BEAM_SWORD_DASH_SWING = 66
    HOME_RUN_BAT_JAB = 67
    HOME_RUN_BAT_TILT_SWING = 68
    HOME_RUN_BAT_SMASH_SWING = 69
    HOME_RUN_BAT_DASH_SWING = 70
    PARASOL_JAB = 71
    PARASOL_TILT_SWING = 72
    PARASOL_SMASH_SWING = 73
    PARASOL_DASH_SWING = 74
    FAN_JAB = 75
    FAN_TILT_SWING = 76
    FAN_SMASH_SWING = 77
    FAN_DASH_SWING = 78
    STAR_ROD_JAB = 79
    STAR_ROD_TILT_SWING = 80
    STAR_ROD_SMASH_SWING = 81
    STAR_ROD_DASH_SWING = 82
    LIPS_STICK_JAB = 83
    LIPS_STICK_TILT_SWING = 84
    LIPS_STICK_SMASH_SWING = 85
    LIPS_STICK_DASH_SWING = 86
    OPEN_PARASOL = 87
    RAY_GUN_SHOOT = 88
    FIRE_FLOWER_SHOOT = 89
    SCREW_ATTACK = 90
    SUPER_SCOPE_RAPID = 91
    SUPER_SCOPE_CHARGED = 92
    HAMMER = 93


class Triggers(Base):
    __slots__ = 'logical', 'physical'

    logical: float #: Processed analog trigger position
    physical: Triggers.Physical #: Physical analog trigger positions (useful for APM)

    def __init__(self, logical: float, physical_x: float, physical_y: float):
        self.logical = logical
        self.physical = self.Physical(physical_x, physical_y)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return other.logical == self.logical and other.physical == self.physical


    class Physical(Base):
        __slots__ = 'l', 'r'

        l: float
        r: float

        def __init__(self, l: float, r: float):
            self.l = l
            self.r = r

        def __eq__(self, other):
            if not isinstance(other, self.__class__):
                return NotImplemented
            # Should we add an epsilon to these comparisons? When are people going to be comparing trigger states for equality, other than in our tests?
            return other.l == self.l and other.r == self.r


class Buttons(Base):
    __slots__ = 'logical', 'physical'

    logical: Buttons.Logical #: Processed button-state bitmask
    physical: Buttons.Physical #: Physical button-state bitmask

    def __init__(self, logical, physical):
        self.logical = self.Logical(logical)
        self.physical = self.Physical(physical)

    def __eq__(self, other):
        if not isinstance(other, Buttons):
            return NotImplemented
        return other.logical is self.logical and other.physical is self.physical


    class Logical(IntFlag):
        TRIGGER_ANALOG = 2**31
        CSTICK_RIGHT = 2**23
        CSTICK_LEFT = 2**22
        CSTICK_DOWN = 2**21
        CSTICK_UP = 2**20
        JOYSTICK_RIGHT = 2**19
        JOYSTICK_LEFT = 2**18
        JOYSTICK_DOWN = 2**17
        JOYSTICK_UP = 2**16
        START = 2**12
        Y = 2**11
        X = 2**10
        B = 2**9
        A = 2**8
        L = 2**6
        R = 2**5
        Z = 2**4
        DPAD_UP = 2**3
        DPAD_DOWN = 2**2
        DPAD_RIGHT = 2**1
        DPAD_LEFT = 2**0
        NONE = 0


    class Physical(IntFlag):
        START = 2**12
        Y = 2**11
        X = 2**10
        B = 2**9
        A = 2**8
        L = 2**6
        R = 2**5
        Z = 2**4
        DPAD_UP = 2**3
        DPAD_DOWN = 2**2
        DPAD_RIGHT = 2**1
        DPAD_LEFT = 2**0
        NONE = 0

        def pressed(self):
            """Returns a list of all buttons being pressed."""
            pressed = []
            for b in self.__class__:
                if self & b:
                    pressed.append(b)
            return pressed


class LCancel(IntEnum):
    SUCCESS = 1
    FAILURE = 2


class StateFlags(IntFlag):
    REFLECT = 2**4
    UNTOUCHABLE = 2**10
    FAST_FALL = 2**11
    HIT_LAG = 2**13
    SHIELD = 2**23
    HIT_STUN = 2**25
    SHIELD_TOUCH = 2**26
    POWER_SHIELD = 2**29
    FOLLOWER = 2**35
    SLEEP = 2**36
    DEAD = 2**38
    OFF_SCREEN = 2**39



# These IDs (and other very useful info for this project) came from the SSBM Data Sheet: https://docs.google.com/spreadsheets/d/1JX2w-r2fuvWuNgGb6D3Cs4wHQKLFegZe2jhbBuIhCG8

from .util import *


class ActionState(IntEnum):
    DEAD_DOWN = 0
    DEAD_LEFT = 1
    DEAD_RIGHT = 2
    DEAD_UP = 3
    DEAD_UP_STAR = 4
    DEAD_UP_STAR_ICE = 5
    DEAD_UP_FALL = 6
    DEAD_UP_FALL_HIT_CAMERA = 7
    DEAD_UP_FALL_HIT_CAMERA_FLAT = 8
    DEAD_UP_FALL_ICE = 9
    DEAD_UP_FALL_HIT_CAMERA_ICE = 10
    SLEEP = 11
    REBIRTH = 12
    REBIRTH_WAIT = 13
    WAIT = 14
    WALK_SLOW = 15
    WALK_MIDDLE = 16
    WALK_FAST = 17
    TURN = 18
    TURN_RUN = 19
    DASH = 20
    RUN = 21
    RUN_DIRECT = 22
    RUN_BRAKE = 23
    KNEE_BEND = 24
    JUMP_F = 25
    JUMP_B = 26
    JUMP_AERIAL_F = 27
    JUMP_AERIAL_B = 28
    FALL = 29
    FALL_F = 30
    FALL_B = 31
    FALL_AERIAL = 32
    FALL_AERIAL_F = 33
    FALL_AERIAL_B = 34
    FALL_SPECIAL = 35
    FALL_SPECIAL_F = 36
    FALL_SPECIAL_B = 37
    DAMAGE_FALL = 38
    SQUAT = 39
    SQUAT_WAIT = 40
    SQUAT_RV = 41
    LANDING = 42
    LANDING_FALL_SPECIAL = 43
    ATTACK_11 = 44
    ATTACK_12 = 45
    ATTACK_13 = 46
    ATTACK_100_START = 47
    ATTACK_100_LOOP = 48
    ATTACK_100_END = 49
    ATTACK_DASH = 50
    ATTACK_S_3_HI = 51
    ATTACK_S_3_HI_S = 52
    ATTACK_S_3_S = 53
    ATTACK_S_3_LW_S = 54
    ATTACK_S_3_LW = 55
    ATTACK_HI_3 = 56
    ATTACK_LW_3 = 57
    ATTACK_S_4_HI = 58
    ATTACK_S_4_HI_S = 59
    ATTACK_S_4_S = 60
    ATTACK_S_4_LW_S = 61
    ATTACK_S_4_LW = 62
    ATTACK_HI_4 = 63
    ATTACK_LW_4 = 64
    ATTACK_AIR_N = 65
    ATTACK_AIR_F = 66
    ATTACK_AIR_B = 67
    ATTACK_AIR_HI = 68
    ATTACK_AIR_LW = 69
    LANDING_AIR_N = 70
    LANDING_AIR_F = 71
    LANDING_AIR_B = 72
    LANDING_AIR_HI = 73
    LANDING_AIR_LW = 74
    DAMAGE_HI_1 = 75
    DAMAGE_HI_2 = 76
    DAMAGE_HI_3 = 77
    DAMAGE_N_1 = 78
    DAMAGE_N_2 = 79
    DAMAGE_N_3 = 80
    DAMAGE_LW_1 = 81
    DAMAGE_LW_2 = 82
    DAMAGE_LW_3 = 83
    DAMAGE_AIR_1 = 84
    DAMAGE_AIR_2 = 85
    DAMAGE_AIR_3 = 86
    DAMAGE_FLY_HI = 87
    DAMAGE_FLY_N = 88
    DAMAGE_FLY_LW = 89
    DAMAGE_FLY_TOP = 90
    DAMAGE_FLY_ROLL = 91
    LIGHT_GET = 92
    HEAVY_GET = 93
    LIGHT_THROW_F = 94
    LIGHT_THROW_B = 95
    LIGHT_THROW_HI = 96
    LIGHT_THROW_LW = 97
    LIGHT_THROW_DASH = 98
    LIGHT_THROW_DROP = 99
    LIGHT_THROW_AIR_F = 100
    LIGHT_THROW_AIR_B = 101
    LIGHT_THROW_AIR_HI = 102
    LIGHT_THROW_AIR_LW = 103
    HEAVY_THROW_F = 104
    HEAVY_THROW_B = 105
    HEAVY_THROW_HI = 106
    HEAVY_THROW_LW = 107
    LIGHT_THROW_F_4 = 108
    LIGHT_THROW_B_4 = 109
    LIGHT_THROW_HI_4 = 110
    LIGHT_THROW_LW_4 = 111
    LIGHT_THROW_AIR_F_4 = 112
    LIGHT_THROW_AIR_B_4 = 113
    LIGHT_THROW_AIR_HI_4 = 114
    LIGHT_THROW_AIR_LW_4 = 115
    HEAVY_THROW_F_4 = 116
    HEAVY_THROW_B_4 = 117
    HEAVY_THROW_HI_4 = 118
    HEAVY_THROW_LW_4 = 119
    SWORD_SWING_1 = 120
    SWORD_SWING_3 = 121
    SWORD_SWING_4 = 122
    SWORD_SWING_DASH = 123
    BAT_SWING_1 = 124
    BAT_SWING_3 = 125
    BAT_SWING_4 = 126
    BAT_SWING_DASH = 127
    PARASOL_SWING_1 = 128
    PARASOL_SWING_3 = 129
    PARASOL_SWING_4 = 130
    PARASOL_SWING_DASH = 131
    HARISEN_SWING_1 = 132
    HARISEN_SWING_3 = 133
    HARISEN_SWING_4 = 134
    HARISEN_SWING_DASH = 135
    STAR_ROD_SWING_1 = 136
    STAR_ROD_SWING_3 = 137
    STAR_ROD_SWING_4 = 138
    STAR_ROD_SWING_DASH = 139
    LIP_STICK_SWING_1 = 140
    LIP_STICK_SWING_3 = 141
    LIP_STICK_SWING_4 = 142
    LIP_STICK_SWING_DASH = 143
    ITEM_PARASOL_OPEN = 144
    ITEM_PARASOL_FALL = 145
    ITEM_PARASOL_FALL_SPECIAL = 146
    ITEM_PARASOL_DAMAGE_FALL = 147
    L_GUN_SHOOT = 148
    L_GUN_SHOOT_AIR = 149
    L_GUN_SHOOT_EMPTY = 150
    L_GUN_SHOOT_AIR_EMPTY = 151
    FIRE_FLOWER_SHOOT = 152
    FIRE_FLOWER_SHOOT_AIR = 153
    ITEM_SCREW = 154
    ITEM_SCREW_AIR = 155
    DAMAGE_SCREW = 156
    DAMAGE_SCREW_AIR = 157
    ITEM_SCOPE_START = 158
    ITEM_SCOPE_RAPID = 159
    ITEM_SCOPE_FIRE = 160
    ITEM_SCOPE_END = 161
    ITEM_SCOPE_AIR_START = 162
    ITEM_SCOPE_AIR_RAPID = 163
    ITEM_SCOPE_AIR_FIRE = 164
    ITEM_SCOPE_AIR_END = 165
    ITEM_SCOPE_START_EMPTY = 166
    ITEM_SCOPE_RAPID_EMPTY = 167
    ITEM_SCOPE_FIRE_EMPTY = 168
    ITEM_SCOPE_END_EMPTY = 169
    ITEM_SCOPE_AIR_START_EMPTY = 170
    ITEM_SCOPE_AIR_RAPID_EMPTY = 171
    ITEM_SCOPE_AIR_FIRE_EMPTY = 172
    ITEM_SCOPE_AIR_END_EMPTY = 173
    LIFT_WAIT = 174
    LIFT_WALK_1 = 175
    LIFT_WALK_2 = 176
    LIFT_TURN = 177
    GUARD_ON = 178
    GUARD = 179
    GUARD_OFF = 180
    GUARD_SET_OFF = 181
    GUARD_REFLECT = 182
    DOWN_BOUND_U = 183
    DOWN_WAIT_U = 184
    DOWN_DAMAGE_U = 185
    DOWN_STAND_U = 186
    DOWN_ATTACK_U = 187
    DOWN_FOWARD_U = 188
    DOWN_BACK_U = 189
    DOWN_SPOT_U = 190
    DOWN_BOUND_D = 191
    DOWN_WAIT_D = 192
    DOWN_DAMAGE_D = 193
    DOWN_STAND_D = 194
    DOWN_ATTACK_D = 195
    DOWN_FOWARD_D = 196
    DOWN_BACK_D = 197
    DOWN_SPOT_D = 198
    PASSIVE = 199
    PASSIVE_STAND_F = 200
    PASSIVE_STAND_B = 201
    PASSIVE_WALL = 202
    PASSIVE_WALL_JUMP = 203
    PASSIVE_CEIL = 204
    SHIELD_BREAK_FLY = 205
    SHIELD_BREAK_FALL = 206
    SHIELD_BREAK_DOWN_U = 207
    SHIELD_BREAK_DOWN_D = 208
    SHIELD_BREAK_STAND_U = 209
    SHIELD_BREAK_STAND_D = 210
    FURA_FURA = 211
    CATCH = 212
    CATCH_PULL = 213
    CATCH_DASH = 214
    CATCH_DASH_PULL = 215
    CATCH_WAIT = 216
    CATCH_ATTACK = 217
    CATCH_CUT = 218
    THROW_F = 219
    THROW_B = 220
    THROW_HI = 221
    THROW_LW = 222
    CAPTURE_PULLED_HI = 223
    CAPTURE_WAIT_HI = 224
    CAPTURE_DAMAGE_HI = 225
    CAPTURE_PULLED_LW = 226
    CAPTURE_WAIT_LW = 227
    CAPTURE_DAMAGE_LW = 228
    CAPTURE_CUT = 229
    CAPTURE_JUMP = 230
    CAPTURE_NECK = 231
    CAPTURE_FOOT = 232
    ESCAPE_F = 233
    ESCAPE_B = 234
    ESCAPE = 235
    ESCAPE_AIR = 236
    REBOUND_STOP = 237
    REBOUND = 238
    THROWN_F = 239
    THROWN_B = 240
    THROWN_HI = 241
    THROWN_LW = 242
    THROWN_LW_WOMEN = 243
    PASS = 244
    OTTOTTO = 245
    OTTOTTO_WAIT = 246
    FLY_REFLECT_WALL = 247
    FLY_REFLECT_CEIL = 248
    STOP_WALL = 249
    STOP_CEIL = 250
    MISS_FOOT = 251
    CLIFF_CATCH = 252
    CLIFF_WAIT = 253
    CLIFF_CLIMB_SLOW = 254
    CLIFF_CLIMB_QUICK = 255
    CLIFF_ATTACK_SLOW = 256
    CLIFF_ATTACK_QUICK = 257
    CLIFF_ESCAPE_SLOW = 258
    CLIFF_ESCAPE_QUICK = 259
    CLIFF_JUMP_SLOW_1 = 260
    CLIFF_JUMP_SLOW_2 = 261
    CLIFF_JUMP_QUICK_1 = 262
    CLIFF_JUMP_QUICK_2 = 263
    APPEAL_R = 264
    APPEAL_L = 265
    SHOULDERED_WAIT = 266
    SHOULDERED_WALK_SLOW = 267
    SHOULDERED_WALK_MIDDLE = 268
    SHOULDERED_WALK_FAST = 269
    SHOULDERED_TURN = 270
    THROWN_F_F = 271
    THROWN_F_B = 272
    THROWN_F_HI = 273
    THROWN_F_LW = 274
    CAPTURE_CAPTAIN = 275
    CAPTURE_YOSHI = 276
    YOSHI_EGG = 277
    CAPTURE_KOOPA = 278
    CAPTURE_DAMAGE_KOOPA = 279
    CAPTURE_WAIT_KOOPA = 280
    THROWN_KOOPA_F = 281
    THROWN_KOOPA_B = 282
    CAPTURE_KOOPA_AIR = 283
    CAPTURE_DAMAGE_KOOPA_AIR = 284
    CAPTURE_WAIT_KOOPA_AIR = 285
    THROWN_KOOPA_AIR_F = 286
    THROWN_KOOPA_AIR_B = 287
    CAPTURE_KIRBY = 288
    CAPTURE_WAIT_KIRBY = 289
    THROWN_KIRBY_STAR = 290
    THROWN_COPY_STAR = 291
    THROWN_KIRBY = 292
    BARREL_WAIT = 293
    BURY = 294
    BURY_WAIT = 295
    BURY_JUMP = 296
    DAMAGE_SONG = 297
    DAMAGE_SONG_WAIT = 298
    DAMAGE_SONG_RV = 299
    DAMAGE_BIND = 300
    CAPTURE_MEWTWO = 301
    CAPTURE_MEWTWO_AIR = 302
    THROWN_MEWTWO = 303
    THROWN_MEWTWO_AIR = 304
    WARP_STAR_JUMP = 305
    WARP_STAR_FALL = 306
    HAMMER_WAIT = 307
    HAMMER_WALK = 308
    HAMMER_TURN = 309
    HAMMER_KNEE_BEND = 310
    HAMMER_FALL = 311
    HAMMER_JUMP = 312
    HAMMER_LANDING = 313
    KINOKO_GIANT_START = 314
    KINOKO_GIANT_START_AIR = 315
    KINOKO_GIANT_END = 316
    KINOKO_GIANT_END_AIR = 317
    KINOKO_SMALL_START = 318
    KINOKO_SMALL_START_AIR = 319
    KINOKO_SMALL_END = 320
    KINOKO_SMALL_END_AIR = 321
    ENTRY = 322
    ENTRY_START = 323
    ENTRY_END = 324
    DAMAGE_ICE = 325
    DAMAGE_ICE_JUMP = 326
    CAPTURE_MASTER_HAND = 327
    CAPTURE_DAMAGE_MASTER_HAND = 328
    CAPTURE_WAIT_MASTER_HAND = 329
    THROWN_MASTER_HAND = 330
    CAPTURE_KIRBY_YOSHI = 331
    KIRBY_YOSHI_EGG = 332
    CAPTURE_REDEAD = 333
    CAPTURE_LIKE_LIKE = 334
    DOWN_REFLECT = 335
    CAPTURE_CRAZY_HAND = 336
    CAPTURE_DAMAGE_CRAZY_HAND = 337
    CAPTURE_WAIT_CRAZY_HAND = 338
    THROWN_CRAZY_HAND = 339
    BARREL_CANNON_WAIT = 340
    WAIT_1 = 341
    WAIT_2 = 342
    WAIT_3 = 343
    WAIT_4 = 344
    WAIT_ITEM = 345
    SQUAT_WAIT_1 = 346
    SQUAT_WAIT_2 = 347
    SQUAT_WAIT_ITEM = 348
    GUARD_DAMAGE = 349
    ESCAPE_N = 350
    ATTACK_S_4_HOLD = 351
    HEAVY_WALK_1 = 352
    HEAVY_WALK_2 = 353
    ITEM_HAMMER_WAIT = 354
    ITEM_HAMMER_MOVE = 355
    ITEM_BLIND = 356
    DAMAGE_ELEC = 357
    FURA_SLEEP_START = 358
    FURA_SLEEP_LOOP = 359
    FURA_SLEEP_END = 360
    WALL_DAMAGE = 361
    CLIFF_WAIT_1 = 362
    CLIFF_WAIT_2 = 363
    SLIP_DOWN = 364
    SLIP = 365
    SLIP_TURN = 366
    SLIP_DASH = 367
    SLIP_WAIT = 368
    SLIP_STAND = 369
    SLIP_ATTACK = 370
    SLIP_ESCAPE_F = 371
    SLIP_ESCAPE_B = 372
    APPEAL_S = 373
    ZITABATA = 374
    CAPTURE_KOOPA_HIT = 375
    THROWN_KOOPA_END_F = 376
    THROWN_KOOPA_END_B = 377
    CAPTURE_KOOPA_AIR_HIT = 378
    THROWN_KOOPA_AIR_END_F = 379
    THROWN_KOOPA_AIR_END_B = 380
    THROWN_KIRBY_DRINK_S_SHOT = 381
    THROWN_KIRBY_SPIT_S_SHOT = 382


class CSSCharacter(IntEnum):
    CAPTAIN_FALCON = 0
    DONKEY_KONG = 1
    FOX = 2
    GAME_AND_WATCH = 3
    KIRBY = 4
    BOWSER = 5
    LINK = 6
    LUIGI = 7
    MARIO = 8
    MARTH = 9
    MEWTWO = 10
    NESS = 11
    PEACH = 12
    PIKACHU = 13
    ICE_CLIMBERS = 14
    JIGGLYPUFF = 15
    SAMUS = 16
    YOSHI = 17
    ZELDA = 18
    SHEIK = 19
    FALCO = 20
    YOUNG_LINK = 21
    DR_MARIO = 22
    ROY = 23
    PICHU = 24
    GANONDORF = 25
    MASTER_HAND = 26
    WIREFRAME_MALE = 27
    WIREFRAME_FEMALE = 28
    GIGA_BOWSER = 29
    CRAZY_HAND = 30
    SANDBAG = 31
    POPO = 32

    @classmethod
    def from_internal_id(cls, internal_id):
        char = InGameCharacter(internal_id)
        if char is InGameCharacter.NANA or char is InGameCharacter.POPO:
            return cls.ICE_CLIMBERS
        else:
            return cls[char.name]


class InGameCharacter(IntEnum):
    MARIO = 0
    FOX = 1
    CAPTAIN_FALCON = 2
    DONKEY_KONG = 3
    KIRBY = 4
    BOWSER = 5
    LINK = 6
    SHEIK = 7
    NESS = 8
    PEACH = 9
    POPO = 10
    NANA = 11
    PIKACHU = 12
    SAMUS = 13
    YOSHI = 14
    JIGGLYPUFF = 15
    MEWTWO = 16
    LUIGI = 17
    MARTH = 18
    ZELDA = 19
    YOUNG_LINK = 20
    DR_MARIO = 21
    FALCO = 22
    PICHU = 23
    GAME_AND_WATCH = 24
    GANONDORF = 25
    ROY = 26
    MASTER_HAND = 27
    CRAZY_HAND = 28
    WIREFRAME_MALE = 29
    WIREFRAME_FEMALE = 30
    GIGA_BOWSER = 31
    SANDBAG = 32


class Stage(IntEnum):
    FOUNTAIN_OF_DREAMS = 2
    POKEMON_STADIUM = 3
    PRINCESS_PEACHS_CASTLE = 4
    KONGO_JUNGLE = 5
    BRINSTAR = 6
    CORNERIA = 7
    YOSHIS_STORY = 8
    ONETT = 9
    MUTE_CITY = 10
    RAINBOW_CRUISE = 11
    JUNGLE_JAPES = 12
    GREAT_BAY = 13
    HYRULE_TEMPLE = 14
    BRINSTAR_DEPTHS = 15
    YOSHIS_ISLAND = 16
    GREEN_GREENS = 17
    FOURSIDE = 18
    MUSHROOM_KINGDOM_I = 19
    MUSHROOM_KINGDOM_II = 20
    VENOM = 22
    POKE_FLOATS = 23
    BIG_BLUE = 24
    ICICLE_MOUNTAIN = 25
    ICETOP = 26
    FLAT_ZONE = 27
    DREAM_LAND_N64 = 28
    YOSHIS_ISLAND_N64 = 29
    KONGO_JUNGLE_N64 = 30
    BATTLEFIELD = 31
    FINAL_DESTINATION = 32


class Item(IntEnum):
    CAPSULE = 0x00
    BOX = 0x01
    BARREL = 0x02
    EGG = 0x03
    PARTY_BALL = 0x04
    BARREL_CANNON = 0x05
    BOB_OMB = 0x06
    MR_SATURN = 0x07
    HEART_CONTAINER = 0x08
    MAXIM_TOMATO = 0x09
    STARMAN = 0x0A
    HOME_RUN_BAT = 0x0B
    BEAM_SWORD = 0x0C
    PARASOL = 0x0D
    GREEN_SHELL_1 = 0x0E
    RED_SHELL_1 = 0x0F
    RAY_GUN = 0x10
    FREEZIE = 0x11
    FOOD = 0x12
    PROXIMITY_MINE = 0x13
    FLIPPER = 0x14
    SUPER_SCOPE = 0x15
    STAR_ROD = 0x16
    LIP_STICK = 0x17
    FAN = 0x18
    FIRE_FLOWER = 0x19
    SUPER_MUSHROOM = 0x1A
    WARP_STAR = 0x1D
    SCREW_ATTACK = 0x1E
    BUNNY_HOOD = 0x1F
    METAL_BOX = 0x20
    CLOAKING_DEVICE = 0x21
    POKE_BALL = 0x22

    # ITEM RELATED
    RAY_GUN_RECOIL_EFFECT = 0x23
    STAR_ROD_STAR = 0x24
    LIP_STICK_DUST = 0x25
    SUPER_SCOPE_BEAM = 0x26
    RAY_GUN_BEAM = 0x27
    HAMMER_HEAD = 0x28
    FLOWER = 0x29
    YOSHI_EGG_1 = 0x2A

    # MONSTERS
    GOOMBA = 0x2B
    REDEAD = 0x2C
    OCTAROK = 0x2D
    OTTOSEA = 0x2E
    STONE = 0x2F

    # CHARACTER RELATED
    MARIO_FIRE = 0x30
    DR_MARIO_PILL = 0x31
    KIRBY_CUTTER_BEAM = 0x32
    KIRBY_HAMMER = 0x33
    FOX_LASER = 0x36
    FALCO_LASER = 0x37
    FOX_SHADOW = 0x38
    FALCO_SHADOW = 0x39
    LINK_BOMB = 0x3A
    YOUNG_LINK_BOMB = 0x3B
    LINK_BOOMERANG = 0x3C
    YOUNG_LINK_BOOMERANG = 0x3D
    LINK_HOOKSHOT = 0x3E
    YOUNG_LINK_HOOKSHOT = 0x3F
    LINK_ARROW_1 = 0x40
    YOUNG_LINK_FIRE_ARROW = 0x41
    NESS_PK_FIRE = 0x42
    NESS_PK_FLASH_1 = 0x43
    NESS_PK_FLASH_2 = 0x44
    NESS_PK_THUNDER_1 = 0x45
    NESS_PK_THUNDER_2 = 0x46
    NESS_PK_THUNDER_3 = 0x47
    NESS_PK_THUNDER_4 = 0x48
    NESS_PK_THUNDER_5 = 0x49
    FOX_BLASTER = 0x4A
    FALCO_BLASTER = 0x4B
    LINK_ARROW_2 = 0x4C
    YOUNG_LINK_ARROW = 0x4D
    NESS_PK_FLASH_3 = 0x4E
    SHEIK_NEEDLE_1 = 0x4F
    SHEIK_NEEDLE_2 = 0x50
    PIKACHU_THUNDER_1 = 0x51
    PICHU_THUNDER_1 = 0x52
    MARIO_CAPE = 0x53
    DR_MARIO_CAPE = 0x54
    SHEIK_SMOKE = 0x55
    YOSHI_EGG_2 = 0x56
    YOSHI_TONGUE_1 = 0x57
    YOSHI_STAR = 0x58
    PIKACHU_THUNDER_2 = 0x59
    PIKACHU_THUNDER_3 = 0x5A
    PICHU_THUNDER_2 = 0x5B
    PICHU_THUNDER_3 = 0x5C
    SAMUS_BOMB = 0x5D
    SAMUS_CHARGESHOT = 0x5E
    SAMUS_MISSILE = 0x5F
    SAMUS_GRAPPLE_BEAM = 0x60
    SHEIK_CHAIN = 0x61
    PEACH_TURNIP = 0x63
    BOWSER_FLAME = 0x64
    NESS_BAT = 0x65
    NESS_YOYO = 0x66
    PEACH_PARASOL = 0x67
    PEACH_TOAD = 0x68
    LUIGI_FIRE = 0x69
    ICE_CLIMBERS_ICE = 0x6A
    ICE_CLIMBERS_BLIZZARD = 0x6B
    ZELDA_FIRE_1 = 0x6C
    ZELDA_FIRE_2 = 0x6D
    PEACH_TOAD_SPORE = 0x6F
    MEWTWO_SHADOWBALL = 0x70
    ICE_CLIMBERS_UP_B = 0x71
    GAME_AND_WATCH_PESTICIDE = 0x72
    GAME_AND_WATCH_MANHOLE = 0x73
    GAME_AND_WATCH_FIRE = 0x74
    GAME_AND_WATCH_PARACHUTE = 0x75
    GAME_AND_WATCH_TURTLE = 0x76
    GAME_AND_WATCH_SPERKY = 0x77
    GAME_AND_WATCH_JUDGE = 0x78
    GAME_AND_WATCH_SAUSAGE = 0x7A
    GAME_AND_WATCH_MILK = 0x7B
    GAME_AND_WATCH_FIREFIGHTER = 0x7C
    MASTER_HAND_LASER = 0x7D
    MASTER_HAND_BULLET = 0x7E
    CRAZY_HAND_LASER = 0x7F
    CRAZY_HAND_BULLET = 0x80
    CRAZY_HAND_BOMB = 0x81
    KIRBY_COPY_MARIO_FIRE = 0x82
    KIRBY_COPY_DR_MARIO_PILL = 0x83
    KIRBY_COPY_LUIGI_FIRE = 0x84
    KIRBY_COPY_ICE_CLIMBERS_ICE = 0x85
    KIRBY_COPY_PEACH_TOAD = 0x86
    KIRBY_COPY_TOAD_SPORE = 0x87
    KIRBY_COPY_FOX_LASER = 0x88
    KIRBY_COPY_FALCO_LASER = 0x89
    KIRBY_COPY_FOX_BLASTER = 0x8A
    KIRBY_COPY_FALCO_BLASTER = 0x8B
    KIRBY_COPY_LINK_ARROW_1 = 0x8C
    KIRBY_COPY_YOUNG_LINK_ARROW_1 = 0x8D
    KIRBY_COPY_LINK_ARROW_2 = 0x8E
    KIRBY_COPY_YOUNG_LINK_ARROW_2 = 0x8F
    KIRBY_COPY_MEWTWO_SHADOWBALL = 0x90
    KIRBY_COPY_PK_FLASH = 0x91
    KIRBY_COPY_PK_FLASH_EXPLOSION = 0x92
    KIRBY_COPY_PIKACHU_THUNDER_1 = 0x93
    KIRBY_COPY_PIKACHU_THUNDER_2 = 0x94
    KIRBY_COPY_PICHU_THUNDER_1 = 0x95
    KIRBY_COPY_PICHU_THUNDER_2 = 0x96
    KIRBY_COPY_SAMUS_CHARGESHOT = 0x97
    KIRBY_COPY_SHEIK_NEEDLE_1 = 0x98
    KIRBY_COPY_SHEIK_NEEDLE_2 = 0x99
    KIRBY_COPY_BOWSER_FLAME = 0x9A
    KIRBY_COPY_GAME_AND_WATCH_SAUSAGE = 0x9B
    YOSHI_TONGUE_2 = 0x9D
    MARIO_LUIGI_COIN = 0x9F

    # POKEMON
    RANDOM_POKEMON = 0xA0
    GOLDEEN = 0xA1
    CHICORITA = 0xA2
    SNORLAX = 0xA3
    BLASTOISE = 0xA4
    WEEZING = 0xA5
    CHARIZARD = 0xA6
    MOLTRES = 0xA7
    ZAPDOS = 0xA8
    ARTICUNO = 0xA9
    WOBBUFFET = 0xAA
    SCIZOR = 0xAB
    UNOWN = 0xAC
    ENTEI = 0xAD
    RAIKOU = 0xAE
    SUICUNE = 0xAF
    BELLOSSOM = 0xB0
    ELECTRODE = 0xB1
    LUGIA = 0xB2
    HO_OH = 0xB3
    DITTO = 0xB4
    CLEFAIRY = 0xB5
    TOGEPI = 0xB6
    MEW = 0xB7
    CELEBI = 0xB8
    STARYU = 0xB9
    CHANSEY = 0xBA
    PORYGON = 0xBB
    CYNDAQUIL = 0xBC
    MARILL = 0xBD
    VENUSAUR = 0xBE

    # POKEMON RELATED
    CHICORITA_LEAF = 0xBF
    BLASTOISE_WATER = 0xC0
    WEEZING_GAS_1 = 0xC1
    WEEZING_GAS_2 = 0xC2
    CHARIZARD_BREATH_1 = 0xC3
    CHARIZARD_BREATH_2 = 0xC4
    CHARIZARD_BREATH_3 = 0xC5
    CHARIZARD_BREATH_4 = 0xC6
    MINI_UNOWNS = 0xC7
    LUGIA_AEROBLAST_1 = 0xC8
    LUGIA_AEROBLAST_2 = 0xC9
    LUGIA_AEROBLAST_3 = 0xCA
    HO_OH_FLAME = 0xCB
    STARYU_STAR = 0xCC
    HEALING_EGG = 0xCD
    CYNDAQUIL_FIRE = 0xCE

    # MONSTERS
    OLD_GOOMBA = 0xD0
    TARGET = 0xD1
    SHYGUY = 0xD2
    KOOPA_1 = 0xD3
    KOOPA_2 = 0xD4
    LIKE_LIKE = 0xD5
    OLD_OTTOSEA = 0xD8
    WHITE_BEAR = 0xD9
    KLAP = 0xDA
    GREEN_SHELL_2 = 0xDB
    RED_SHELL_2 = 0xDC

    # STAGE SPECIFIC
    TINGLE = 0xDD
    APPLE = 0xE1
    HEALING_APPLE = 0xE2
    TOOL = 0xE6
    BIRDO = 0xE9
    ARWING_LASER = 0xEA
    GREAT_FOX_LASER = 0xEB
    BIRDO_EGG = 0xEC
